import time
import threading
import unittest

from pebble.pebble import TimeoutError, SerializingError
from pebble.decorators import asynchronous, concurrent


_task_id = ""
_results = 0
_exception = None


class UnserializeableError(Exception):
    def __init__(self):
        self.lock = threading.Lock()  # unpickleable

    def __str__(self):
        return "BOOM!"


def callback(task_id, results):
    global _task_id
    global _results
    _task_id = task_id
    _results = results


def error_callback(task_id, results):
    global _task_id
    global _exception
    _task_id = task_id
    _exception = results


@asynchronous
def ajob(argument, keyword_argument=0):
    return argument + keyword_argument


@asynchronous
def ajob_error(argument, keyword_argument=0):
    raise Exception("BOOM!")


@asynchronous(callback=callback)
def ajob_callback(argument, keyword_argument=0):
    return argument + keyword_argument


@asynchronous
def ajob_long():
    time.sleep(1)
    return 1


@asynchronous
def ajob_count():
    return None


@concurrent
def cjob(argument, keyword_argument=0):
    return argument + keyword_argument


@concurrent
def cjob_error(argument, keyword_argument=0):
    raise Exception("BOOM!")


@concurrent(callback=callback)
def cjob_callback(argument, keyword_argument=0):
    return argument + keyword_argument


@concurrent
def cjob_long():
    time.sleep(1)
    return 1


@concurrent
def cjob_count():
    return None


@concurrent(timeout=1)
def cjob_timeout():
    time.sleep(2)
    return 1


@concurrent
def cjob_unserializeable():
    return threading.Lock()


@concurrent
def cjob_unserializeable_error():
    raise UnserializeableError()


class TestPebbleDecorators(unittest.TestCase):
    def setUp(self):
        global _task_id
        global _results
        global _exception
        _task_id = ""
        _results = 0
        _exception = None
        self.task_id = ''
        self.exception = None
        self.callback_results = 0

    def callback(self, task_id, results):
        self.task_id = task_id
        self.callback_results = results

    def error_callback(self, task_id, exception):
        self.task_id = task_id
        self.exception = exception

    def test_asynchronous_wrong_decoration(self):
        """Decorator raises ValueError if given wrong params."""
        try:
            @asynchronous(callback, error_callback)
            def wrong(argument, keyword_argument=0):
                return argument + keyword_argument
        except Exception as error:
            self.assertTrue(isinstance(error, ValueError))

    def test_asynchronous_task(self):
        """Test that an asynchronous task is correctly executed."""
        task = ajob(1, 1)
        self.assertEqual(task.get(), 2)

    def test_asynchronous_task_callback_static(self):
        """Test static callback is executed with asynchronous task."""
        task = ajob_callback(1, 1)
        task.get()
        self.assertEqual((task.id, 2), (_task_id, _results))

    def test_asynchronous_task_callback_dynamic(self):
        """Test dynamic callback is executed with asynchronous task."""
        ajob.callback = self.callback
        task = ajob(1, 1)
        task.get()
        self.assertEqual((task.id, 2), (self.task_id, self.callback_results))

    def test_asynchronous_task_error(self):
        """Test that an exception in an asynch task is raised by get()."""
        task = ajob_error(1, 1)
        try:
            task.get()
        except:
            pass
        self.assertRaises(Exception, task.get)

    def test_asynchronous_task_error_callback(self):
        """Test that an exception in a task is managed in error_callback."""
        ajob_error.error_callback = self.error_callback
        task = ajob_error(1, 1)
        try:
            task.get()
        except:
            pass
        self.assertEqual((task.id, 'BOOM!'),
                         (self.task_id, str(self.exception)))

    def test_asynchronous_task_long(self):
        """Test timeout get parameter works."""
        task = ajob_long()
        self.assertEqual(task.get(2), 1)

    def test_asynchronous_timeout_error(self):
        """TimeoutError is raised if task has not yet finished."""
        task = ajob_long()
        self.assertRaises(TimeoutError, task.get, 0)

    def test_asynchronous_task_number(self):
        """Task number are correctly assigned."""
        for i in range(0, 5):
            task = ajob_count(1, 1)
        self.assertEqual(task.number, 4)

    def test_concurrent_wrong_decoration(self):
        """Decorator raises ValueError if given wrong params."""
        try:
            @concurrent(callback, error_callback)
            def wrong(argument, keyword_argument=0):
                return argument + keyword_argument
        except Exception as error:
            self.assertTrue(isinstance(error, ValueError))

    def test_concurrent_task(self):
        """Test that a concurrent task is correctly executed."""
        task = cjob(1, 1)
        self.assertEqual(task.get(), 2)

    def test_concurrent_task_callback_static(self):
        """Test static callback is executed with concurrent task."""
        task = cjob_callback(1, 1)
        task.get()
        self.assertEqual((task.id, 2), (_task_id, _results))

    def test_concurrent_task_callback_dynamic(self):
        """Test dynamic callback is executed with concurrent task."""
        cjob.callback = self.callback
        task = cjob(1, 1)
        task.get()
        self.assertEqual((task.id, 2), (self.task_id, self.callback_results))

    def test_concurrent_task_error(self):
        """Test that an exception in a concurrent task is raised by get()."""
        task = cjob_error(1, 1)
        self.assertRaises(Exception, task.get)

    def test_concurrent_task_error_callback(self):
        """Test that an exception in a task is managed in error_callback."""
        cjob_error.error_callback = self.error_callback
        task = cjob_error(1, 1)
        try:
            task.get()
        except:
            pass
        self.assertEqual((task.id, 'BOOM!'),
                         (self.task_id, str(self.exception)))

    def test_concurrent_task_long(self):
        """Test timeout get parameter works."""
        task = cjob_long()
        self.assertEqual(task.get(2), 1)

    def test_concurrent_timeout_error(self):
        """TimeoutError is raised if task has not yet finished."""
        task = cjob_long()
        self.assertRaises(TimeoutError, task.get, 0)

    def test_concurrent_task_number(self):
        """Task number are correctly assigned."""
        for i in range(0, 5):
            task = cjob_count(1, 1)
        self.assertEqual(task.number, 4)

    def test_concurrent_timeout(self):
        """Timeout decorator kills the task."""
        task = cjob_timeout()
        self.assertFalse(task.get(), None)

    def test_concurrent_timeout_not_expired(self):
        """Timeout decorator doesn't kill the task."""
        cjob_timeout.timeout = 5
        task = cjob_timeout()
        self.assertEqual(task.get(), 1)

    def test_concurrent_unserializable(self):
        """PicklingError is returned if results are not serializeable."""
        task = cjob_unserializeable()
        try:  # Python 2
            from cPickle import PicklingError
            self.assertRaises(PicklingError, task.get)
        except ImportError:  # Python 3
            from pickle import PicklingError
            self.assertRaises(PicklingError, task.get)

    def test_concurrent_unserializable_error(self):
        """SerializingError is returned if exception is not serializeable."""
        task = cjob_unserializeable_error()
        self.assertRaises(SerializingError, task.get)


if __name__ == "__main__":
    unittest.main()



# This file is part of Pebble.

# Pebble is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.

# Pebble is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with Pebble.  If not, see <http://www.gnu.org/licenses/>.


from itertools import count
from threading import Thread
from collections import Callable
from functools import update_wrapper
from multiprocessing import Process, Pipe

from .pebble import ThreadTask, ProcessTask, thread_worker, process_worker


class AsynchronousWrapper(object):
    def __init__(self, function, callback, error_callback):
        self._function = function
        self._counter = count()
        self.callback = callback
        self.error_callback = error_callback
        update_wrapper(self, function)

    def __call__(self, *args, **kwargs):
        t = ThreadTask(next(self._counter),
                       self.callback, self.error_callback)
        args = list(args)
        args.insert(0, self._function)
        args.insert(1, t)
        t._worker = Thread(target=thread_worker, args=(args), kwargs=(kwargs))
        t._worker.daemon = True
        t._worker.start()
        return t


def asynchronous(*args, **kwargs):
    """Turns a *function* into a Thread and runs its logic within.

    A decorated *function* will return a *Task* object once is called.

    If *callback* is a callable, it will be called once the task has ended
    with the task identifier and the *function* return values.
    If *error_callback* is a callable, it will be called if the task has raised
    an exception, passing the task identifier and the raised exception.

    """
    def wrapper(function):
        return AsynchronousWrapper(function, callback, error_callback)

    if len(args) == 1 and not len(kwargs) and isinstance(args[0], Callable):
        return AsynchronousWrapper(args[0], None, None)
    elif not len(args) and len(kwargs):
        callback = kwargs.get('callback', None)
        error_callback = kwargs.get('error_callback', None)

        return wrapper
    else:
        raise ValueError("Decorator accepts only keyword arguments.")


class ConcurrentWrapper(object):
    def __init__(self, function, timeout, callback, error_callback):
        self._function = function
        self._counter = count()
        self.timeout = timeout
        self.callback = callback
        self.error_callback = error_callback
        update_wrapper(self, function)

    def __call__(self, *args, **kwargs):
        reader, writer = Pipe(duplex=False)
        args = list(args)
        args.insert(0, self._function)
        args.insert(1, writer)
        p = Process(target=process_worker, args=(args), kwargs=(kwargs))
        p.daemon = True
        p.start()
        return ProcessTask(next(self._counter), p, reader,
                           self.callback, self.error_callback, self.timeout)


def concurrent(*args, **kwargs):
    """Turns a *function* into a Process and runs its logic within.

    A decorated *function* will return a *Task* object once is called.

    If *callback* is a callable, it will be called once the task has ended
    with the task identifier and the *function* return values.
    If *error_callback* is a callable, it will be called if the task has raised
    an exception, passing the task identifier and the raised exception.

    """
    def wrapper(function):
        return ConcurrentWrapper(function, timeout, callback, error_callback)

    if len(args) == 1 and not len(kwargs) and isinstance(args[0], Callable):
        return ConcurrentWrapper(args[0], None, None, None)
    elif not len(args) and len(kwargs):
        timeout = kwargs.get('timeout', None)
        callback = kwargs.get('callback', None)
        error_callback = kwargs.get('error_callback', None)

        return wrapper
    else:
        raise ValueError("Decorator accepts only keyword arguments.")



# This file is part of Pebble.

# Pebble is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.

# Pebble is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with Pebble.  If not, see <http://www.gnu.org/licenses/>.


from uuid import uuid4
from traceback import format_exc

from threading import Thread
try:  # Python 2
    from cPickle import PicklingError
except:  # Python 3
    from pickle import PicklingError


def thread_worker(function, task, *args, **kwargs):
    try:
        results = function(*args, **kwargs)
        task._set(results)
    except Exception as error:
        error.traceback = format_exc()
        task._set(error)


def process_worker(function, writer, *args, **kwargs):
    try:
        writer.send(function(*args, **kwargs))
    except (IOError, OSError):  # pipe was closed
        return
    except Exception as error:
        error.traceback = format_exc()
        try:
            writer.send(error)
        except PicklingError:
            writer.send(SerializingError(str(error), type(error)))


class PebbleError(Exception):
    """Pebble base exception."""
    pass


class TimeoutError(PebbleError):
    """Raised when Task.get() timeout expires."""
    def __init__(self, msg):
        self.msg = msg

    def __repr__(self):
        return "%s %s" % (self.__class__, self.msg)

    def __str__(self):
        return str(self.msg)


class SerializingError(PebbleError):
    """Raised if unable to serialize an Exception."""
    def __init__(self, msg, value):
        super(SerializingError, self).__init__(msg, value)
        self.msg = msg
        self.value = value

    def __repr__(self):
        return "%s %s: %s\n%s" % (self.__class__, self.value, self.msg)

    def __str__(self):
        return "Unable to serialize %s. Message: %s" % (self.value, self.msg)


class ThreadTask(object):
    def __init__(self, task_nr, callback, error_callback):
        self.id = uuid4()
        self.number = task_nr
        self._results = None
        self._worker = None  # set by Asynchronous._wrapper
        self._callback = callback
        self._error_callback = error_callback

    def get(self, timeout=None):
        self._worker.join(timeout)
        if not self._worker.is_alive():
            if (isinstance(self._results, Exception)):
                raise self._results
            else:
                return self._results
        else:
            raise TimeoutError("Task is still running")

    def _set(self, results):
        self._results = results
        if (isinstance(self._results, Exception) and
                self._error_callback is not None):
            self._error_callback(self.id, self._results)
        elif self._callback is not None:
            self._callback(self.id, self._results)


class ProcessTask(object):
    def __init__(self, task_nr, worker, reader,
                 callback, error_callback, timeout):
        self.id = uuid4()
        self.number = task_nr
        self._timeout = timeout
        self._results = None
        self._reader = reader
        self._worker = worker
        self._callback = callback
        self._error_callback = error_callback
        self._worker_listener = Thread(target=self._set)
        self._worker_listener.daemon = True
        self._worker_listener.start()

    def get(self, timeout=None):
        self._worker_listener.join(timeout)
        if not self._worker_listener.is_alive():
            if (isinstance(self._results, Exception)):
                raise self._results
            else:
                return self._results
        else:
            raise TimeoutError("Task is still running")

    def _set(self):
        try:
            if self._reader.poll(self._timeout):
                self._results = self._reader.recv()
                if (isinstance(self._results, Exception) and
                        self._error_callback is not None):
                    self._error_callback(self.id, self._results)
                elif self._callback is not None:
                    self._callback(self.id, self._results)
            elif self._worker.is_alive():
                self._worker.terminate()
        except (IOError, OSError) as error:  # pipe was closed
            self._results = error
        finally:
            self._worker.join()



import os
from setuptools import setup


# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="Pebble",
    version="1.1.5",
    author="Matteo Cafasso",
    author_email="noxdafox@gmail.com",
    description=("Threading and multiprocessing eye-candy."),
    license="LGPL",
    keywords="thread process pool decorator",
    url="https://github.com/noxdafox/pebble",
    packages=['pebble', 'tests'],
    long_description=read('README.txt'),
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)"
    ],
)



__all__ = ['asynchronous', 'concurrent', 'PebbleError',
           'TimeoutError', 'SerializingError']

from .pebble import PebbleError, TimeoutError, SerializingError
from .decorators import asynchronous, concurrent
